
from ...model cimport Node, Vector, Matrix44
from ...language.vm cimport Function

from libc.stdint cimport int64_t, uint64_t


cdef double DefaultSnapAngle
cdef int64_t DefaultSegments


cpdef tuple build_arrays_from_trimesh(trimesh_model)
cpdef tuple build_arrays_from_manifold(manifold)
cpdef void fill_in_normals(vertices_array, faces_array)


cdef class Model:
    cdef readonly uint64_t id
    cdef readonly double touch_timestamp
    cdef readonly dict cache
    cdef readonly set dependents
    cdef readonly list buffer_caches

    cpdef bint uncache(self, bint buffers)
    cpdef void unload(self)
    cpdef void check_for_changes(self)
    cpdef bint is_manifold(self)
    cpdef double signed_distance(self, double x, double y, double z) noexcept
    cpdef tuple build_arrays(self)
    cpdef object build_trimesh(self)
    cpdef object build_manifold(self)

    cpdef void add_dependent(self, Model model)
    cpdef void remove_dependent(self, Model model)
    cpdef void invalidate(self)
    cpdef Vector get_bounds(self)
    cpdef tuple get_arrays(self)
    cpdef object get_trimesh(self)
    cpdef object get_manifold(self)
    cpdef tuple get_buffers(self, object glctx, dict objects)

    cpdef Model flatten(self)
    cpdef Model invert(self)
    cpdef Model repair(self)
    cdef Model _snap_edges(self, double snap_angle)
    cdef Model _transform(self, Matrix44 transform_matrix)
    cdef Model _uv_remap(self, str mapping)
    cdef Model _trim(self, Vector origin, Vector normal, double smooth, double fillet, double chamfer)

    @staticmethod
    cdef Model _boolean(str operation, list models, double smooth, double fillet, double chamfer)

    @staticmethod
    cdef Model _box(str uv_map)

    @staticmethod
    cdef Model _sphere(int64_t segments)

    @staticmethod
    cdef Model _cylinder(int64_t segments)

    @staticmethod
    cdef Model _cone(int64_t segments)

    @staticmethod
    cdef Model _external(str filename)

    @staticmethod
    cdef Model _vector(Vector vertices, Vector faces)

    @staticmethod
    cdef Model _sdf(Function function, Model original, Vector minimum, Vector maximum, double resolution)

    @staticmethod
    cdef Model _mix(list models, Vector weights)



"""
Flitter language integration tests
"""

from pathlib import Path
import unittest

from flitter import configure_logger
from flitter.language.tree import Literal, Let, Sequence, Export, Binding, Top
from flitter.language.parser import parse, ParseError
from flitter.language.vm import Function
from flitter.model import Vector, Node, Context, StateDict, DummyStateDict, null

from . import utils


configure_logger('WARNING')


class TestPragmas(unittest.TestCase):
    def test_no_pragmas(self):
        top = parse("\n")
        self.assertEqual(top.pragmas, ())
        program = top.compile()
        self.assertEqual(program.pragmas, {})

    def test_pragma_not_at_top(self):
        with self.assertRaises(ParseError):
            parse("""
let x=10
%pragma tempo 120
""")

    def test_pragmas(self):
        top = parse("""
%pragma foo 5
%pragma bar 'five'
%pragma baz :five
%pragma buz !five
""")
        self.assertEqual(repr(top), repr(Top((Binding('foo', Literal(5)),
                                              Binding('bar', Literal('five')),
                                              Binding('baz', Literal(Vector.symbol('five'))),
                                              Binding('buz', Literal(Node('five')))), Sequence((Export(None),)))))
        program = top.compile()
        self.assertEqual(program.pragmas, {'foo': Vector(5), 'bar': Vector('five'), 'baz': Vector.symbol('five'), 'buz': Vector(Node('five'))})


class TestParseErrors(unittest.TestCase):
    def test_missing_bindings(self):
        with self.assertRaises(ParseError):
            parse("""
let
!foo
""")

    def test_unexpected_indent(self):
        with self.assertRaises(ParseError):
            parse("""
let x=5
    !foo x=x
""")

    def test_missing_indent(self):
        with self.assertRaises(ParseError):
            parse("""
for x in ..5
!foo x=x
""")

    def bad_identifier(self):
        with self.assertRaises(ParseError):
            parse("""
let 1x=5
""")

    def mixed_position_and_named_args(self):
        with self.assertRaises(ParseError):
            parse("""
f(1, x=2, 3)
""")


class TestLanguageFeatures(unittest.TestCase):
    """
    Test a series of language features by parsing the given code, evaluating it and comparing the
    output to that given. The code is evaluated in two ways:

    1. Compile the unsimplified AST and run the resulting VM program with any additional bound
       names passed in as initial lvalues
    2. Simplify the AST with any additional bound names passed in as static, then extract the
       literal values out of the resulting AST

    It is assumed that all of the examples can be fully reduced to a literal by the simplifier.
    """

    def assertCodeOutput(self, code, output, with_errors=None, skip_simplifier=False, **names):
        top = parse(code.strip())
        output = output.strip()
        if with_errors is None:
            with_errors = set()
        vm_context = Context(names={name: Vector(value) for name, value in names.items()}, state=StateDict())
        vm_output = '\n'.join(repr(node) for node in top.compile(initial_lnames=tuple(names)).run(vm_context).root.children)
        self.assertEqual(vm_output, output, msg="VM output is incorrect")
        self.assertEqual(vm_context.errors, with_errors)
        if skip_simplifier:
            return
        simplified_top, simplifier_context = top.simplify(static=names, return_context=True)
        self.assertEqual(simplifier_context.errors, with_errors)
        expr = simplified_top.body
        while not isinstance(expr, Literal):
            if isinstance(expr, Let):
                expr = expr.body
            elif isinstance(expr, Sequence) and len(expr.expressions) == 2 and isinstance(expr.expressions[1], Export):
                expr = expr.expressions[0]
            else:
                break
        if isinstance(expr, Export):
            simplifier_output = ""
        else:
            self.assertIsInstance(expr, Literal, msg="Unexpected simplification output")
            simplifier_output = '\n'.join(repr(node) for node in expr.value)
        self.assertEqual(simplifier_output, output, msg="Simplifier output is incorrect")
        mixed_top, simplifier_context = top.simplify(dynamic=names, return_context=True)
        vm_context = Context(names={name: Vector(value) for name, value in names.items()}, state=StateDict())
        vm_output = '\n'.join(repr(node) for node in mixed_top.compile(initial_lnames=tuple(names)).run(vm_context).root.children)
        self.assertEqual(vm_output, output, msg="VM output is incorrect")
        self.assertEqual(simplifier_context.errors | vm_context.errors, with_errors)

    def test_empty(self):
        self.assertCodeOutput("", "")

    def test_literal_node(self):
        self.assertCodeOutput("!foo #bar x=5 y=:baz z='Hello world!'", "!foo #bar x=5 y=:baz z='Hello world!'")

    def test_unicode_names(self):
        self.assertCodeOutput("!café #århus 水=:𓀀", "!café #århus 水=:𓀀")

    def test_names_with_primes(self):
        self.assertCodeOutput(
            """
let y=1 y''=y+1 y'''=y''+1
!node' x'=y''' y=:y'''' z='prime'
            """,
            """
!node' x'=3 y=:y'''' z='prime'
            """)

    def test_named_values(self):
        self.assertCodeOutput(
            """
!node values=null;true;false;inf;-inf;nan
            """,
            """
!node values=1;0;inf;-inf;nan
            """
        )

    def test_let_only(self):
        self.assertCodeOutput("let x=5", "")

    def test_contextual_parser(self):
        self.assertCodeOutput(
            """
let in=import + 1
!foo let=in
            """,
            """
!foo let=6
            """, **{'import': 5})

    def test_names(self):
        self.assertCodeOutput(
            """
!foo x=x y=y z=z #bar
            """,
            """
!foo #bar x=5 y=:baz z='Hello world!'
            """, x=5, y=Vector.symbol('baz'), z='Hello world!')

    def test_let(self):
        self.assertCodeOutput(
            """
let x=5 y=:baz z='Hello world!'
!foo #bar x=x y=y z=z
            """,
            """
!foo #bar x=5 y=:baz z='Hello world!'
            """)

    def test_where(self):
        self.assertCodeOutput(
            """
let z=2
!foo x=y+z where y=z*2 where z=5
            """,
            """
!foo x=15
            """)

    def test_multibind_loop(self):
        self.assertCodeOutput(
            """
for x;y in ..9
    !node x=x y=y z=z
            """,
            """
!node x=0 y=1 z=0
!node x=2 y=3 z=0
!node x=4 y=5 z=0
!node x=6 y=7 z=0
!node x=8 z=0
            """, z=0)

    def test_nested_loops(self):
        self.assertCodeOutput(
            """
let n=4
for i in ..n
    !group i=i
        for j in ..i
            !item numbers=(k*(j+1) for k in m..i+1)
            """,
            """
!group i=0
!group i=1
 !item numbers=1
!group i=2
 !item numbers=1;2
 !item numbers=2;4
!group i=3
 !item numbers=1;2;3
 !item numbers=2;4;6
 !item numbers=3;6;9
            """, m=1)

    def test_if(self):
        self.assertCodeOutput(
            """
for i in ..2
    if i
        !one
            """,
            """
!one
            """)

    def test_if_else(self):
        self.assertCodeOutput(
            """
for i in ..4
    if not i
        !zero
    elif i == 2
        !two
    elif i > 2
        !more
    else
        !one
            """,
            """
!zero
!one
!two
!more
            """)

    def test_recursive_inlined_function(self):
        # Also default parameter values and out-of-order named arguments
        self.assertCodeOutput(
            """
func fib(n, x=0, y=1)
    fib(n-1, y=y+x, x=y) if n > 0 else x

!fib x=fib(10)
            """,
            """
!fib x=55
            """)

    def test_builtin_calls(self):
        self.assertCodeOutput(
            """
for i in ..n
    !color rgb=hsv(i/n;1;1)
            """,
            """
!color rgb=1;0;0
!color rgb=1;1;0
!color rgb=0;1;0
!color rgb=0;1;1
!color rgb=0;0;1
!color rgb=1;0;1
            """, n=6)

    def test_template_calls(self):
        self.assertCodeOutput(
            """
func foo(nodes, x=10, y, z='world')
    !foo x=x z='hello';z
        nodes #test y=y

@foo
@foo z='you'
    !bar
@foo y=5
    @foo z='me' x=99
        !baz
            """,
            """
!foo x=10 z='hello';'world'
!foo x=10 z='hello';'you'
 !bar #test
!foo x=10 z='hello';'world'
 !foo #test x=99 z='hello';'me' y=5
  !baz #test
            """)

    def test_template_builtin_calls(self):
        self.assertCodeOutput(
            """
func foo(n)
    @sum
        for i in ..n
            2**i

!foo foo=foo(10)
            """,
            """
!foo foo=1023
            """)

    def test_nested_functions(self):
        self.assertCodeOutput(
            """
func fib(n)
    func fib'(n, x, y)
        fib'(n-1, y=y+x, x=y) if n > 0 else x
    fib'(n, 0, 1)

!fib x=fib(10)
            """,
            """
!fib x=55
            """)

    def test_anonymous_function(self):
        """Note that this is statically reducible because `map` is inlined and so the
           anonymous function is bound to `f`, which therefore becomes a function name"""
        self.assertCodeOutput(
            """
func map(f, xs)
    for x in xs
        f(x)

!doubled x=map(func(x, y=2) x*y, ..10)
            """,
            """
!doubled x=0;2;4;6;8;10;12;14;16;18
            """)

    def test_anonymous_function_with_where(self):
        self.assertCodeOutput(
            """
let f = func(x) x+y where y=x*x
!foo bar=f(10)
            """,
            """
!foo bar=110
            """
        )

    def test_anonymous_function_returning_anonymous_function(self):
        self.assertCodeOutput(
            """
let f = func(x) func(y) x + y
!foo bar=f(10)(5)
            """,
            """
!foo bar=15
            """
        )

    def test_accidental_anonymous_vector(self):
        self.assertCodeOutput(
            """
let f = func(x) x;1;2
!foo bar=f(0)
            """,
            """
!foo bar=0
            """,
            with_errors={'1.0 is not callable', '2.0 is not callable'},
            skip_simplifier=True
        )

    def test_some_deliberately_obtuse_behaviour(self):
        self.assertCodeOutput(
            """
@(func(_, x) ((!foo x=x*y) for y in ..3)) x=5
            """,
            """
!foo x=0
!foo x=5
!foo x=10
            """)

    def test_bad_function_call(self):
        self.assertCodeOutput(
            """
!foo color=hsv()
            """,
            """
!foo
            """, with_errors={'Error calling hsv: hsv() takes exactly 1 positional argument (0 given)'})

    def test_contains(self):
        self.assertCodeOutput(
            """
!foo x=(x in y) y=(y in y) z=(y in x)
            """,
            """
!foo x=1 y=1 z=0
            """, x=Vector((4, 5, 6)), y=Vector.range(10))

    def test_sequence_let(self):
        self.assertCodeOutput(
            """
let foo;bar =
    !foo x=1
        !bar
        !baz
    !bar b=2

!frob
    bar
    foo
            """,
            """
!frob
 !bar b=2
 !foo x=1
  !bar
  !baz
            """)


class ScriptTest(utils.TestCase):
    """
    Run a Flitter script under three different conditions and test that the generated root nodes
    and exported values match:

    1. Compile the AST and run it without *any* simplification and with an empty state
    2. Simplify the AST with known names (`beat`, etc.) as dynamic and an empty state
    3. Simplify the AST with known names as static values and a special dummy state that claims to
       contain all keys but returns null for all as an empty state would

    In the third case, all state lookups will be statically evaluated to literal nulls by the
    simplifier. The idea is to exercise the simplifier as much as possible.
    """

    def assertExportsMatch(self, exports1, exports2, msg=None):
        self.assertEqual(exports1.keys(), exports2.keys())
        for name in exports1:
            a = exports1[name]
            b = exports2[name]
            if a.isinstance(Function) and b.isinstance(Function):
                self.assertEqual(a[0].__name__, b[0].__name__, msg=msg)
            elif a.numeric and b.numeric:
                self.assertAllAlmostEqual(a, b, msg=msg)
            else:
                self.assertEqual(a, b, msg=msg)

    def assertSimplifierDoesntChangeBehaviour(self, script):
        self.maxDiff = None
        state = StateDict()
        null_state = DummyStateDict()
        names = {'beat': Vector(0), 'tempo': Vector(120), 'quantum': Vector(4), 'fps': Vector(60), 'OUTPUT': null,
                 'run_time': Vector(1), 'frame': Vector(0)}
        top = parse(script.read_text(encoding='utf8'))
        # Compile un-simplified AST and execute program:
        program1 = top.compile(initial_lnames=tuple(names))
        program1.set_path(script)
        program1.use_simplifier(False)
        context = program1.run(Context(names=dict(names), state=state))
        root1 = context.root
        exports1 = context.exports
        windows = [node for node in root1.children if node.kind == 'window']
        self.assertEqual(len(windows), 1, msg="Should be a single window node in program output")
        self.assertGreater(len(tuple(windows[0].children)), 0, "Output window node should have children")
        # Simplify AST with dynamic names, compile and execute:
        top2 = top.simplify(dynamic=set(names))
        self.assertEqual(repr(top2.simplify(dynamic=set(names))), repr(top2), msg="Dynamic simplification not complete in one step")
        program2 = top2.compile(initial_lnames=tuple(names))
        program2.set_path(script)
        self.assertNotEqual(str(program1), str(program2), msg="Dynamically-simplified program should be different from original")
        context = program2.run(Context(names=dict(names), state=state))
        root2 = context.root
        exports2 = context.exports
        self.assertEqual(repr(root2), repr(root1), msg="Dynamically-simplified program output doesn't match original")
        self.assertExportsMatch(exports2, exports1, msg="Dynamically-simplified program exports don't match original")
        # Simplify AST with static names and null state, compile and execute:
        top3 = top.simplify(static=names, state=null_state)
        self.assertIs(top3.simplify(static=names, state=null_state), top3, msg="Static simplification not complete in one step")
        program3 = top3.compile()
        program3.set_path(script)
        self.assertNotEqual(len(program3), len(program1), msg="Statically-simplified program length should be different from original")
        context = program3.run(Context(state=state))
        root3 = context.root
        exports3 = context.exports
        for name in names:
            del exports3[name]
        self.assertEqual(repr(root3), repr(root1), msg="Statically-simplified program output doesn't match original")
        self.assertExportsMatch(exports3, exports1, msg="Statically-simplified program exports don't match original")


class TestDocumentationDiagrams(ScriptTest):
    """
    Run all documentation diagrams with and without simplification.
    """

    DIAGRAMS = Path(__file__).parent.parent / 'docs/diagrams'

    def test_box_uvmap(self):
        self.assertSimplifierDoesntChangeBehaviour(self.DIAGRAMS / 'box_uvmap.fl')

    def test_dummyshader(self):
        self.assertSimplifierDoesntChangeBehaviour(self.DIAGRAMS / 'dummyshader.fl')

    def test_easings(self):
        self.assertSimplifierDoesntChangeBehaviour(self.DIAGRAMS / 'easings.fl')

    def test_pseudorandoms(self):
        self.assertSimplifierDoesntChangeBehaviour(self.DIAGRAMS / 'pseudorandoms.fl')

    def test_spheroidbox(self):
        self.assertSimplifierDoesntChangeBehaviour(self.DIAGRAMS / 'spheroidbox.fl')

    def test_torus(self):
        self.assertSimplifierDoesntChangeBehaviour(self.DIAGRAMS / 'torus.fl')

    def test_waveforms(self):
        self.assertSimplifierDoesntChangeBehaviour(self.DIAGRAMS / 'waveforms.fl')

    def test_petri(self):
        self.assertSimplifierDoesntChangeBehaviour(self.DIAGRAMS / 'petri.fl')


class TestDocumentationTutorial(ScriptTest):
    """
    Run all tutorial images with and without simplification.
    """

    TUTORIAL_IMAGES = Path(__file__).parent.parent / 'docs/tutorial_images'

    def test_tutorial1(self):
        self.assertSimplifierDoesntChangeBehaviour(self.TUTORIAL_IMAGES / 'tutorial1.fl')

    def test_tutorial2(self):
        self.assertSimplifierDoesntChangeBehaviour(self.TUTORIAL_IMAGES / 'tutorial2.fl')

    def test_tutorial3(self):
        self.assertSimplifierDoesntChangeBehaviour(self.TUTORIAL_IMAGES / 'tutorial3.fl')

    def test_tutorial4(self):
        self.assertSimplifierDoesntChangeBehaviour(self.TUTORIAL_IMAGES / 'tutorial4.fl')

    def test_tutorial5(self):
        self.assertSimplifierDoesntChangeBehaviour(self.TUTORIAL_IMAGES / 'tutorial5.fl')

    def test_tutorial6(self):
        self.assertSimplifierDoesntChangeBehaviour(self.TUTORIAL_IMAGES / 'tutorial6.fl')


class TestExamples(ScriptTest):
    """
    Run all examples with and without simplification.
    """

    EXAMPLES = Path(__file__).parent.parent / 'examples'

    def test_bauble(self):
        self.assertSimplifierDoesntChangeBehaviour(self.EXAMPLES / 'bauble.fl')

    def test_bounce(self):
        self.assertSimplifierDoesntChangeBehaviour(self.EXAMPLES / 'bounce.fl')

    def test_canvas3d(self):
        self.assertSimplifierDoesntChangeBehaviour(self.EXAMPLES / 'canvas3d.fl')

    def test_dots(self):
        self.assertSimplifierDoesntChangeBehaviour(self.EXAMPLES / 'dots.fl')

    def test_hoops(self):
        self.assertSimplifierDoesntChangeBehaviour(self.EXAMPLES / 'hoops.fl')

    def test_linear(self):
        self.assertSimplifierDoesntChangeBehaviour(self.EXAMPLES / 'linear.fl')

    def test_linelight(self):
        self.assertSimplifierDoesntChangeBehaviour(self.EXAMPLES / 'linelight.fl')

    def test_oklch(self):
        self.assertSimplifierDoesntChangeBehaviour(self.EXAMPLES / 'oklch.fl')

    def test_physics(self):
        self.assertSimplifierDoesntChangeBehaviour(self.EXAMPLES / 'physics.fl')

    def test_sdf(self):
        self.assertSimplifierDoesntChangeBehaviour(self.EXAMPLES / 'sdf.fl')

    def test_smoke(self):
        self.assertSimplifierDoesntChangeBehaviour(self.EXAMPLES / 'smoke.fl')

    def test_solidgeometry(self):
        self.assertSimplifierDoesntChangeBehaviour(self.EXAMPLES / 'solidgeometry.fl')

    def test_sphere(self):
        self.assertSimplifierDoesntChangeBehaviour(self.EXAMPLES / 'sphere.fl')

    def test_teaset(self):
        self.assertSimplifierDoesntChangeBehaviour(self.EXAMPLES / 'teaset.fl')

    def test_textures(self):
        self.assertSimplifierDoesntChangeBehaviour(self.EXAMPLES / 'textures.fl')

    def test_translucency(self):
        self.assertSimplifierDoesntChangeBehaviour(self.EXAMPLES / 'translucency.fl')

    def test_video(self):
        self.assertSimplifierDoesntChangeBehaviour(self.EXAMPLES / 'video.fl')




from cpython cimport PyObject

from libc.stdint cimport int64_t, uint64_t

from ..model cimport Vector, Node, Context


cdef dict static_builtins
cdef dict dynamic_builtins
cdef dict builtins


cdef class VectorStack:
    cdef PyObject** vectors
    cdef int64_t top
    cdef readonly int64_t size
    cdef readonly int64_t max_size

    cpdef VectorStack copy(self)
    cpdef void drop(self, int64_t count=?)
    cpdef void push(self, Vector vector)
    cpdef Vector pop(self)
    cpdef tuple pop_tuple(self, int64_t count)
    cpdef list pop_list(self, int64_t count)
    cpdef dict pop_dict(self, tuple keys)
    cpdef Vector pop_composed(self, int64_t count)
    cpdef Vector peek(self)
    cpdef Vector peek_at(self, int64_t offset)
    cpdef void poke(self, Vector vector)
    cpdef void poke_at(self, int64_t offset, Vector vector)


cdef class Function:
    cdef readonly str __name__
    cdef readonly Vector vself
    cdef readonly tuple parameters
    cdef readonly tuple defaults
    cdef readonly Program program
    cdef readonly int64_t address
    cdef readonly bint record_stats
    cdef readonly tuple captures
    cdef readonly int64_t call_depth
    cdef readonly uint64_t _hash

    cdef uint64_t hash(self)
    cdef Vector call_one_fast(self, Context context, Vector arg)


cdef enum OpCode:
    Add
    Append
    Attributes
    BeginFor
    BranchFalse
    BranchTrue
    Call
    CallFast
    Ceil
    Compose
    Contains
    Drop
    Dup
    EndFor
    Eq
    Floor
    FloorDiv
    Fract
    Func
    Exit
    Ge
    Gt
    Import
    IndexLiteral
    Jump
    Label
    Le
    Literal
    LiteralNode
    LiteralNodes
    LocalDrop
    LocalLoad
    LocalPush
    Lookup
    LookupLiteral
    Lt
    Mod
    Mul
    MulAdd
    Ne
    Neg
    Next
    Not
    Pos
    Pow
    Range
    Slice
    SliceLiteral
    Export
    Sub
    Tag
    TrueDiv
    Xor
    MAX


cdef class Instruction:
    cdef readonly OpCode code


cdef class InstructionVector(Instruction):
    cdef readonly Vector value


cdef class InstructionNode(Instruction):
    cdef readonly Node value


cdef class InstructionStr(Instruction):
    cdef readonly str value


cdef class InstructionTuple(Instruction):
    cdef readonly tuple value


cdef class InstructionInt(Instruction):
    cdef readonly int64_t value


cdef class InstructionIntTuple(Instruction):
    cdef readonly int64_t ivalue
    cdef readonly tuple tvalue


cdef class InstructionLabel(Instruction):
    cdef readonly int64_t label


cdef class InstructionJump(Instruction):
    cdef readonly int64_t label
    cdef readonly int64_t offset


cdef class InstructionFunc(InstructionJump):
    cdef readonly str name
    cdef readonly tuple parameters
    cdef readonly int64_t ncaptures


cdef class InstructionObjectInt(Instruction):
    cdef readonly object obj
    cdef readonly int64_t value


cdef class Program:
    cdef readonly dict pragmas
    cdef readonly list instructions
    cdef bint linked
    cdef readonly object path
    cdef readonly object top
    cdef readonly tuple initial_lnames
    cdef readonly VectorStack stack
    cdef readonly VectorStack lnames
    cdef readonly set compiler_errors
    cdef readonly bint simplify
    cdef int64_t next_label

    cdef Instruction last_instruction(self)
    cdef Instruction pop_instruction(self)
    cdef Program push_instruction(self, Instruction instruction)
    cpdef Program link(self)
    cpdef void use_simplifier(self, bint simplify)
    cpdef int64_t new_label(self)
    cpdef Program dup(self)
    cpdef Program drop(self, int64_t count=?)
    cpdef Program label(self, int64_t label)
    cpdef Program jump(self, int64_t label)
    cpdef Program branch_true(self, int64_t label)
    cpdef Program branch_false(self, int64_t label)
    cpdef Program import_(self, tuple names)
    cpdef Program literal(self, value)
    cpdef Program local_push(self, int64_t count)
    cpdef Program local_load(self, int64_t offset)
    cpdef Program local_drop(self, int64_t count)
    cpdef Program lookup(self)
    cpdef Program lookup_literal(self, Vector value)
    cpdef Program range(self)
    cpdef Program neg(self)
    cpdef Program pos(self)
    cpdef Program ceil(self)
    cpdef Program floor(self)
    cpdef Program fract(self)
    cpdef Program not_(self)
    cpdef Program add(self)
    cpdef Program sub(self)
    cpdef Program mul(self)
    cpdef Program mul_add(self)
    cpdef Program truediv(self)
    cpdef Program floordiv(self)
    cpdef Program mod(self)
    cpdef Program pow(self)
    cpdef Program contains(self)
    cpdef Program eq(self)
    cpdef Program ne(self)
    cpdef Program gt(self)
    cpdef Program lt(self)
    cpdef Program ge(self)
    cpdef Program le(self)
    cpdef Program xor(self)
    cpdef Program slice(self)
    cpdef Program slice_literal(self, Vector value)
    cpdef Program call(self, int64_t count, tuple names=?)
    cpdef Program call_fast(self, function, int64_t count)
    cpdef Program tag(self, str name)
    cpdef Program attributes(self, tuple names)
    cpdef Program append(self, int64_t count=?)
    cpdef Program compose(self, int64_t count)
    cpdef Program begin_for(self, int64_t count)
    cpdef Program next(self, int64_t label)
    cpdef Program end_for(self)
    cpdef Program export(self, str name)
    cpdef Program func(self, int64_t label, str name, tuple parameters, int64_t ncaptures=?)
    cpdef Program exit(self)
    cdef void _execute(self, Context context, int64_t pc, bint record_stats)



"""
Video window node
"""

from av.video.reformatter import VideoReformatter

from . import ProgramNode
from ...cache import SharedCache
from .glconstants import GL_SRGB8_ALPHA8
from .glsl import TemplateLoader


Reformatter = VideoReformatter()


class Video(ProgramNode):
    DEFAULT_VERTEX_SOURCE = TemplateLoader.get_template('video.vert')
    DEFAULT_FRAGMENT_SOURCE = TemplateLoader.get_template('video.frag')

    def __init__(self, glctx):
        super().__init__(glctx)
        self._filename = None
        self._threading = None
        self._frame0 = None
        self._frame1 = None
        self._frame0_texture = None
        self._frame1_texture = None

    def release(self):
        if self._filename is not None:
            SharedCache[self._filename].read_video_frames(self, None, threading=self._threading)
            self._filename = None
            self._threading = None
        self._frame0_texture = None
        self._frame1_texture = None
        self._frame0 = None
        self._frame1 = None
        super().release()

    @property
    def child_textures(self):
        return {'frame0': self._frame0_texture, 'frame1': self._frame1_texture}

    def similar_to(self, node):
        return super().similar_to(node) and node.get('filename', 1, str) == self._filename

    async def descend(self, engine, node, references, **kwargs):
        # A video is a leaf node from the perspective of the OpenGL world
        pass

    def render(self, node, references, **kwargs):
        self._filename = node.get('filename', 1, str)
        position = node.get('position', 1, float, 0)
        loop = node.get('loop', 1, bool, False)
        back_and_forth = node.get('back_and_forth', 1, bool, False)
        trim_start, trim_end = node.get('trim', 2, float, (0, 0))
        self._threading = node.get('thread', 1, bool, False)
        aspect = {'fit': 1, 'fill': 2}.get(node.get('aspect', 1, str), 0)
        if self._filename is not None:
            ratio, frame0, frame1 = SharedCache[self._filename].read_video_frames(self, position, loop=loop, back_and_forth=back_and_forth,
                                                                                  trim_start=trim_start, trim_end=trim_end, threading=self._threading)
        else:
            ratio, frame0, frame1 = 0, None, None
        if self._frame0_texture is not None and (frame0 is None or (frame0.width, frame0.height) != self._frame0_texture.size):
            self._frame0_texture = None
            self._frame1_texture = None
            self._frame0 = None
            self._frame1 = None
        if frame0 is None:
            if self._target is not None:
                self._target.release()
                self._target = None
            return
        frame_size = frame0.width, frame0.height
        if self._frame0_texture is None:
            self._frame0_texture = self.glctx.texture(frame_size, 4, internal_format=GL_SRGB8_ALPHA8)
            self._frame1_texture = self.glctx.texture(frame_size, 4, internal_format=GL_SRGB8_ALPHA8)
        if frame0 is self._frame1 or frame1 is self._frame0:
            self._frame0_texture, self._frame1_texture = self._frame1_texture, self._frame0_texture
            self._frame0, self._frame1 = self._frame1, self._frame0
        if frame0 is not self._frame0:
            rgb_frame = Reformatter.reformat(frame0, format='rgba')
            plane = rgb_frame.planes[0]
            data = rgb_frame.to_ndarray().data if plane.line_size > plane.width * 4 else plane
            self._frame0_texture.write(memoryview(data))
            self._frame0 = frame0
        if frame1 is None:
            self._frame1 = None
        elif frame1 is not self._frame1:
            rgb_frame = Reformatter.reformat(frame1, format='rgba')
            plane = rgb_frame.planes[0]
            data = rgb_frame.to_ndarray().data if plane.line_size > plane.width * 4 else plane
            self._frame1_texture.write(memoryview(data))
            self._frame1 = frame1
        interpolate = node.get('interpolate', 1, bool, False)
        super().render(node, references, frame_size=frame_size, aspect_mode=aspect, ratio=ratio if interpolate else 0, **kwargs)



"""
Record window node
"""

from . import ProgramNode
from ...cache import SharedCache


class Record(ProgramNode):
    def render(self, node, references, *, time=None, fps=None, realtime=None, **kwargs):
        super().render(node, references, srgb=True, colorbits=8, **kwargs)
        if filename := node.get('filename', 1, str):
            path = SharedCache[filename]
            ext = path.suffix.lower()
            codec = node.get('codec', 1, str, 'h264')
            keep_alpha = node.get('keep_alpha', 1, bool, False)
            if ext in ('.mp4', '.mov', '.m4v', '.mkv', '.webm', '.ogg') or (ext == '.gif' and codec == 'gif'):
                pixfmt = node.get('pixfmt', 1, str, 'rgb8' if codec == 'gif' else ('yuva420p' if keep_alpha else 'yuv420p'))
                limit = node.get('limit', 1, float)
                options = {}
                for key, value in node.items():
                    if key not in {'filename', 'codec', 'pixfmt', 'limit', 'hidden'}:
                        options[key.replace('__', '-')] = str(value)
                path.write_video_frame(self._target.video_frame, time, fps=int(fps), realtime=realtime, codec=codec,
                                       pixfmt=pixfmt, options=options, limit=limit, alpha=keep_alpha)
            else:
                quality = node.get('quality', 1, int)
                path.write_image(self._target.image, quality=quality, alpha=keep_alpha)



"""
Flitter parser tests
"""

import unittest

from flitter.language.tree import (
    Literal, Name, Top, Export, Sequence,
    Positive, Negative, Power, Add, Subtract, Multiply, Divide, FloorDivide, Modulo,
    # Contains, EqualTo, NotEqualTo, LessThan, GreaterThan, LessThanOrEqualTo, GreaterThanOrEqualTo,
    # Not, And, Or, Xor,
    # Range, Slice, Lookup,
    # Tag, Attributes, Append,
    # Let, Call, For, IfElse,
    # Import, Function,
    # Binding, PolyBinding, IfCondition
)
from flitter.model import Vector, Node
from flitter.language.parser import parse, ParseError


class ParserTestCase(unittest.TestCase):
    def assertParsesTo(self, code, expression):
        self.maxDiff = 1000
        if not isinstance(expression, Top):
            if not isinstance(expression, Sequence):
                expression = Sequence((expression, Export(None)))
            elif not isinstance(expression.expressions[-1], Export):
                expression = Sequence(expression.expressions + (Export(None),))
            expression = Top((), expression)
        self.assertEqual(repr(parse(code)), repr(expression))


class TestAtoms(ParserTestCase):
    def test_numbers(self):
        self.assertParsesTo("0", Literal(0))
        self.assertParsesTo("1234567890", Literal(1234567890))
        self.assertParsesTo("0123456789", Literal(123456789))
        self.assertParsesTo("1_234_567_890", Literal(1234567890))
        self.assertParsesTo("1.234567890", Literal(1.23456789))
        self.assertParsesTo("1.234_567_890", Literal(1.23456789))
        self.assertParsesTo("1e9", Literal(1000000000))
        self.assertParsesTo("1.23456789e9", Literal(1234567890))
        self.assertParsesTo("1.234_567_890e9", Literal(1234567890))
        self.assertParsesTo("123.456_789e9", Literal(123456789000))
        self.assertParsesTo("1.23456789e-9", Literal(1.23456789e-9))
        self.assertParsesTo("1.23456789E-99", Literal(1.23456789e-99))
        self.assertParsesTo("1.2_3_4_5_6_7_8_9E1_0_0_0", Literal(1.23456789e1000))

    def test_timecodes(self):
        self.assertParsesTo("0:0", Literal(0))
        self.assertParsesTo("0:0.0", Literal(0))
        self.assertParsesTo("0:0:0", Literal(0))
        self.assertParsesTo("0:0:0.0", Literal(0))
        self.assertParsesTo("000:00:00.000", Literal(0))
        self.assertParsesTo("0:1", Literal(1))
        self.assertParsesTo("000:00:01.000", Literal(1))
        self.assertParsesTo("0:0.1", Literal(0.1))
        self.assertParsesTo("000:00:00.100", Literal(0.1))
        self.assertParsesTo("1:0", Literal(60))
        self.assertParsesTo("000:01:00.000", Literal(60))
        self.assertParsesTo("1:0:0", Literal(3600))
        self.assertParsesTo("001:00:00.000", Literal(3600))
        self.assertParsesTo("1:1:1", Literal(3661))
        self.assertParsesTo("001:01:01.000", Literal(3661))
        self.assertParsesTo("001:01:01.100", Literal(3661.1))
        self.assertParsesTo("999:59:59.999", Literal(3599999.999))
        with self.assertRaises(ParseError):
            parse("0:0:0:0")
        with self.assertRaises(ParseError):
            parse("0:60")
        with self.assertRaises(ParseError):
            parse("60:00")

    def test_si_prefixes(self):
        self.assertParsesTo('1T', Literal(1e12))
        self.assertParsesTo('1G', Literal(1e9))
        self.assertParsesTo('1M', Literal(1e6))
        self.assertParsesTo('1k', Literal(1e3))
        self.assertParsesTo('1m', Literal(1e-3))
        self.assertParsesTo('1u', Literal(1e-6))
        self.assertParsesTo('1µ', Literal(1e-6))
        self.assertParsesTo('1n', Literal(1e-9))
        self.assertParsesTo('1p', Literal(1e-12))

    def test_strings(self):
        self.assertParsesTo('"Hello world!"', Literal("Hello world!"))
        self.assertParsesTo("'Hello world!'", Literal("Hello world!"))
        self.assertParsesTo('''"Hello 'world!'"''', Literal("Hello 'world!'"))
        self.assertParsesTo("""'Hello "world!"'""", Literal('Hello "world!"'))
        self.assertParsesTo('''"""Hello
world!"""''', Literal("Hello\nworld!"))
        self.assertParsesTo("""'''Hello
world!'''""", Literal("Hello\nworld!"))
        self.assertParsesTo("'Hello\\nworld!'", Literal("Hello\nworld!"))
        self.assertParsesTo("'Hello Hafnarfjörður!'", Literal("Hello Hafnarfjörður!"))

    def test_symbols(self):
        self.assertParsesTo(":hello", Literal(Vector.symbol("hello")))
        self.assertParsesTo(":_world", Literal(Vector.symbol("_world")))
        self.assertParsesTo(":hello_world", Literal(Vector.symbol("hello_world")))
        self.assertParsesTo(":_", Literal(Vector.symbol("_")))
        self.assertParsesTo(":Hafnarfjörður", Literal(Vector.symbol("Hafnarfjörður")))
        self.assertParsesTo(":hello123", Literal(Vector.symbol("hello123")))

    def test_nodes(self):
        self.assertParsesTo("!hello", Literal(Node("hello")))
        self.assertParsesTo("!_world", Literal(Node("_world")))
        self.assertParsesTo("!hello_world", Literal(Node("hello_world")))
        self.assertParsesTo("!_", Literal(Node("_")))
        self.assertParsesTo("!Hafnarfjörður", Literal(Node("Hafnarfjörður")))
        self.assertParsesTo("!hello123", Literal(Node("hello123")))

    def test_names(self):
        self.assertParsesTo("hello", Name("hello"))
        self.assertParsesTo("_world", Name("_world"))
        self.assertParsesTo("hello_world", Name("hello_world"))
        self.assertParsesTo("_", Name("_"))
        self.assertParsesTo("Hafnarfjörður", Name("Hafnarfjörður"))
        self.assertParsesTo("hello123", Name("hello123"))


class TestPrecedence(ParserTestCase):
    def test_maths_literals(self):
        self.assertParsesTo("0-1//2%3++4*-5**6/7",
                            Add(Subtract(Literal(0), Modulo(FloorDivide(Literal(1), Literal(2)), Literal(3))),
                                Divide(Multiply(Positive(Literal(4)), Negative(Power(Literal(5), Literal(6)))), Literal(7))))

    def test_maths_names(self):
        self.assertParsesTo("a-b//c%d++e*-f**g/h",
                            Add(Subtract(Name('a'), Modulo(FloorDivide(Name('b'), Name('c')), Name('d'))),
                                Divide(Multiply(Positive(Name('e')), Negative(Power(Name('f'), Name('g')))), Name('h'))))

    def test_maths_parenthesised(self):
        self.assertParsesTo("(a-b)//c%d++e*(-f)**(g/h)",
                            Add(Modulo(FloorDivide(Subtract(Name('a'), Name('b')), Name('c')), Name('d')),
                                Multiply(Positive(Name('e')), Power(Negative(Name('f')), Divide(Name('g'), Name('h'))))))



"""
Flitter plugins
"""

import importlib.metadata

from loguru import logger


PluginEntryPoints = {}
PluginObjectCache = {}


def get_plugin(group, name, quiet=False):
    if group in PluginObjectCache:
        cache = PluginObjectCache[group]
        if name in cache:
            return cache[name]
    else:
        cache = {}
        PluginObjectCache[group] = cache
    if group in PluginEntryPoints:
        entry_points = PluginEntryPoints[group]
    else:
        entry_points = importlib.metadata.entry_points(group=group)
        if entry_points.names:
            logger.debug("Available {} plugins: {}", group, ', '.join(sorted(entry_points.names)))
        PluginEntryPoints[group] = entry_points
    if name in entry_points.names:
        try:
            plugin = entry_points[name].load()
        except Exception:
            logger.exception("Unable to load {} plugin: {}", group, name)
            plugin = None
    else:
        if not quiet:
            logger.warning("No {} plugin found for: {}", group, name)
        plugin = None
    cache[name] = plugin
    return plugin


def get_plugin_names(group):
    if group in PluginEntryPoints:
        entry_points = PluginEntryPoints[group]
    else:
        entry_points = importlib.metadata.entry_points(group=group)
        logger.debug("Available {} plugins: {}", group, ', '.join(entry_points.names))
        PluginEntryPoints[group] = entry_points
    return entry_points.names




import multiprocessing
import os
from pathlib import Path
from setuptools import setup, Extension

from Cython.Build import cythonize


def find_cython_files(dirpath):
    for filepath in dirpath.iterdir():
        if filepath.is_dir():
            yield from find_cython_files(filepath)
        elif filepath.suffix == '.pyx':
            yield filepath


if __name__ == '__main__':
    if int(os.environ.get('FLITTER_BUILD_COVERAGE', '0')):
        print("** Building for coverage testing **")
        define_macros = [("CYTHON_TRACE_NOGIL", "1"), ("CYTHON_USE_SYS_MONITORING", "0")]
        compiler_directives = {'linetrace': True}
        annotate = False
    elif int(os.environ.get('FLITTER_BUILD_PROFILE', '0')):
        print("** Building with profiling support **")
        define_macros = []
        compiler_directives = {'profile': True}
        annotate = True
    else:
        define_macros = []
        compiler_directives = {}
        annotate = False
    ext_modules = [Extension('.'.join(filepath.with_suffix('').parts[1:]), [str(filepath)], define_macros=define_macros)
                   for filepath in find_cython_files(Path('src'))]
    ext_modules = cythonize(ext_modules, language_level=3, compiler_directives=compiler_directives, annotate=annotate,
                            nthreads=multiprocessing.cpu_count())
    setup(ext_modules=ext_modules)



"""
Tests of flitter 3D models
"""

import math
import unittest.mock

import numpy as np
import trimesh

from flitter.model import Vector, Matrix44
from flitter.render.window.models import Model

from . import utils


DefaultSegments = 64


class TestPrimitives(utils.TestCase):
    def tearDown(self):
        Model.flush_caches(0, 0)

    def test_box(self):
        model = Model.box()
        self.assertFalse(model.is_manifold())
        mesh = model.get_trimesh()
        self.assertEqual(model.name, '!box')
        self.assertEqual(mesh.bounds.tolist(), [[-0.5, -0.5, -0.5], [0.5, 0.5, 0.5]])
        self.assertEqual(len(mesh.vertices), 24)
        self.assertEqual(len(mesh.faces), 12)
        self.assertEqual(mesh.area, 6)
        self.assertEqual(mesh.volume, 1)
        model = Model.box('repeat')
        self.assertEqual(model.name, '!box-repeat')

    def test_sphere(self):
        for segments in (4, DefaultSegments, 1024):
            with self.subTest(segments=segments):
                model = Model.sphere(segments)
                self.assertFalse(model.is_manifold())
                self.assertEqual(model.name, f'!sphere-{segments}' if segments != DefaultSegments else '!sphere')
                mesh = model.get_trimesh()
                self.assertEqual(mesh.bounds.tolist(), [[-1, -1, -1], [1, 1, 1]])
                nrows = segments // 4
                self.assertEqual(len(mesh.vertices), 4*(nrows+1)*(nrows+2))
                self.assertEqual(len(mesh.faces), 8*nrows*nrows)
                if segments == 4:
                    self.assertAlmostEqual(mesh.area, 4*math.sqrt(3))
                    self.assertAlmostEqual(mesh.volume, 4/3)
                else:
                    self.assertAlmostEqual(mesh.area, 4*math.pi, places=int(math.log10(segments)))
                    self.assertAlmostEqual(mesh.volume, 4/3*math.pi, places=int(math.log10(segments)))
        self.assertEqual(Model.sphere(0).name, '!sphere-4')
        self.assertEqual(Model.sphere(5).name, '!sphere-8')
        self.assertIs(Model.sphere(), Model.sphere(DefaultSegments))

    def test_cylinder(self):
        for segments in (4, DefaultSegments, 1024):
            with self.subTest(segments=segments):
                model = Model.cylinder(segments)
                self.assertFalse(model.is_manifold())
                self.assertEqual(model.name, f'!cylinder-{segments}' if segments != DefaultSegments else '!cylinder')
                mesh = model.get_trimesh()
                self.assertEqual(mesh.bounds.tolist(), [[-1, -1, -0.5], [1, 1, 0.5]])
                self.assertEqual(len(mesh.vertices), 6*(segments+1))
                self.assertEqual(len(mesh.faces), 4*segments)
                if segments == 4:
                    self.assertAlmostEqual(mesh.area, 4*(1+math.sqrt(2)))
                    self.assertAlmostEqual(mesh.volume, 2)
                else:
                    self.assertAlmostEqual(mesh.area, 4*math.pi, places=int(math.log10(segments)))
                    self.assertAlmostEqual(mesh.volume, math.pi, places=int(math.log10(segments)))
        self.assertEqual(Model.cylinder(0).name, '!cylinder-2')
        self.assertIs(Model.cylinder(), Model.cylinder(DefaultSegments))

    def test_cone(self):
        for segments in (4, DefaultSegments, 1024):
            with self.subTest(segments=segments):
                model = Model.cone(segments)
                self.assertFalse(model.is_manifold())
                self.assertEqual(model.name, f'!cone-{segments}' if segments != DefaultSegments else '!cone')
                mesh = model.get_trimesh()
                self.assertEqual(mesh.bounds.tolist(), [[-1, -1, -0.5], [1, 1, 0.5]])
                self.assertEqual(len(mesh.vertices), 4*(segments+1))
                self.assertEqual(len(mesh.faces), 2*segments)
                if segments == 4:
                    self.assertAlmostEqual(mesh.area, 2*(1+math.sqrt(3)))
                    self.assertAlmostEqual(mesh.volume, 2/3)
                else:
                    self.assertAlmostEqual(mesh.area, (1+math.sqrt(2))*math.pi, places=int(math.log10(segments)))
                    self.assertAlmostEqual(mesh.volume, math.pi/3, places=int(math.log10(segments)))
        self.assertEqual(Model.cone(0).name, '!cone-2')
        self.assertIs(Model.cone(), Model.cone(DefaultSegments))


class TestBasicFunctionality(utils.TestCase):
    def tearDown(self):
        Model.flush_caches(0, 0)

    def test_equality(self):
        self.assertTrue(Model.box() == Model.box())
        self.assertFalse(Model.box() == Model.sphere())

    def test_str(self):
        self.assertEqual(str(Model.sphere(128)), '!sphere-128')

    def test_repr(self):
        self.assertEqual(repr(Model.sphere(128)), '<Sphere(0x935c42374d93e335)>')

    def test_by_id(self):
        self.assertIsNone(Model.by_id(0x935c42374d93e335))
        model = Model.sphere(128)
        self.assertIs(Model.by_id(0x935c42374d93e335), model)


class MyModel(Model):
    @staticmethod
    def get():
        model = Model.by_id(123)
        if model is None:
            model = MyModel(123)
        return model

    @property
    def name(self):
        return '!mymodel'

    def is_manifold(self):
        return False

    def check_for_changes(self):
        pass

    def build_arrays(self):
        return np.array([[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0]], dtype='f4'), np.array([[0, 1, 2], [2, 3, 0]], dtype='i4')


class TestSubclassing(utils.TestCase):
    def tearDown(self):
        Model.flush_caches(0, 0)

    def test_subclass_instantiation(self):
        model = MyModel.get()
        self.assertIsInstance(model, MyModel)
        self.assertEqual(model.name, '!mymodel')
        self.assertEqual(model.id, 123)
        self.assertIs(MyModel.get(), model)
        mesh = model.get_trimesh()
        self.assertIs(model.get_trimesh(), mesh)
        self.assertAllAlmostEqual(mesh.vertices, [[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0]])
        self.assertAllAlmostEqual(mesh.faces, [[0, 1, 2], [2, 3, 0]])
        self.assertEqual(model.snap_edges().name, 'snap_edges(!mymodel)')

    def test_bad_manifold(self):
        model = MyModel.get()
        with unittest.mock.patch('flitter.render.window.models.logger') as mock_logger:
            manifold = model.get_manifold()
            mock_logger.error.assert_called_with("Mesh is not a volume: {}", model.name)
        self.assertIsNone(manifold)


class TestVector(utils.TestCase):
    def test_bad(self):
        self.assertIsNone(Model.vector(None, None))
        self.assertIsNone(Model.vector('hello', 1))
        self.assertIsNone(Model.vector(1, 'hello'))

    def test_basic(self):
        vertices = [0, 0, 0,  1, 0, 0,  0, 1, 0,  0, 0, 1,  1, 0, 1,  0, 1, 1]
        faces = [0, 1, 2,  3, 4, 5,  0, 1, 4,  4, 3, 0,  1, 2, 5,  5, 4, 1,  2, 0, 3,  3, 5, 2]
        model = Model.vector(vertices, faces)
        self.assertIs(model, Model.vector(vertices, faces))
        self.assertIsNot(model, Model.vector(vertices, reversed(faces)))
        mesh = model.get_trimesh()
        self.assertEqual(mesh.bounds.tolist(), [[0, 0, 0], [1, 1, 1]])
        self.assertAllAlmostEqual(mesh.vertices, np.array(vertices).reshape((-1, 3)))
        self.assertAllAlmostEqual(mesh.faces, np.array(faces).reshape((-1, 3)))
        self.assertAlmostEqual(mesh.area, 3+math.sqrt(2))
        self.assertAlmostEqual(mesh.volume, 0.5)

    def test_invalid(self):
        with unittest.mock.patch('flitter.render.window.models.logger') as mock_logger:
            model = Model.vector([0, 0, 0, 0, 0, 1, 0, 1], [0, 1, 2])
            self.assertIsNone(model.get_trimesh())
            mock_logger.error.assert_called_with("Bad vertices vector length: {}", model.name)
        with unittest.mock.patch('flitter.render.window.models.logger') as mock_logger:
            model = Model.vector([0, 0, 0, 0, 0, 1, 0, 1, 0], [0, 1])
            self.assertIsNone(model.get_trimesh())
            mock_logger.error.assert_called_with("Bad faces vector length: {}", model.name)
        with unittest.mock.patch('flitter.render.window.models.logger') as mock_logger:
            model = Model.vector([0, 0, 0, 0, 0, 1, 0, 1, 0], [0, 1, 3])
            self.assertIsNone(model.get_trimesh())
            mock_logger.error.assert_called_with("Bad vertex index: {}", model.name)


class TestManifoldPrimitives(utils.TestCase):
    def tearDown(self):
        Model.flush_caches(0, 0)

    def test_box(self):
        mesh = Model.box().get_manifold().to_mesh()
        mesh = trimesh.Trimesh(vertices=mesh.vert_properties, faces=mesh.tri_verts, process=False)
        self.assertEqual(mesh.bounds.tolist(), [[-0.5, -0.5, -0.5], [0.5, 0.5, 0.5]])
        self.assertEqual(len(mesh.vertices), 8)
        self.assertEqual(len(mesh.faces), 12)
        self.assertEqual(mesh.area, 6)
        self.assertEqual(mesh.volume, 1)

    def test_sphere(self):
        for segments in (4, DefaultSegments, 1024):
            with self.subTest(segments=segments):
                mesh = Model.sphere(segments).get_manifold().to_mesh()
                mesh = trimesh.Trimesh(vertices=mesh.vert_properties, faces=mesh.tri_verts, process=False)
                self.assertEqual(mesh.bounds.tolist(), [[-1, -1, -1], [1, 1, 1]])
                nrows = segments // 2
                self.assertEqual(len(mesh.vertices), 2+nrows*nrows)
                self.assertEqual(len(mesh.faces), 2*nrows*nrows)
                if segments == 4:
                    self.assertAlmostEqual(mesh.area, 4*math.sqrt(3))
                    self.assertAlmostEqual(mesh.volume, 4/3)
                else:
                    self.assertAlmostEqual(mesh.area, 4*math.pi, places=int(math.log10(segments)))
                    self.assertAlmostEqual(mesh.volume, 4/3*math.pi, places=int(math.log10(segments)))

    def test_cylinder(self):
        for segments in (4, DefaultSegments, 1024):
            with self.subTest(segments=segments):
                mesh = Model.cylinder(segments).get_manifold().to_mesh()
                mesh = trimesh.Trimesh(vertices=mesh.vert_properties, faces=mesh.tri_verts, process=False)
                self.assertEqual(mesh.bounds.tolist(), [[-1, -1, -0.5], [1, 1, 0.5]])
                self.assertEqual(len(mesh.vertices), 2*segments)
                self.assertEqual(len(mesh.faces), 4*(segments-1))
                if segments == 4:
                    self.assertAlmostEqual(mesh.area, 4*(1+math.sqrt(2)))
                    self.assertAlmostEqual(mesh.volume, 2)
                else:
                    self.assertAlmostEqual(mesh.area, 4*math.pi, places=int(math.log10(segments)))
                    self.assertAlmostEqual(mesh.volume, math.pi, places=int(math.log10(segments)))

    def test_cone(self):
        for segments in (4, DefaultSegments, 1024):
            with self.subTest(segments=segments):
                mesh = Model.cone(segments).get_manifold().to_mesh()
                mesh = trimesh.Trimesh(vertices=mesh.vert_properties, faces=mesh.tri_verts, process=False)
                self.assertEqual(mesh.bounds.tolist(), [[-1, -1, -0.5], [1, 1, 0.5]])
                self.assertEqual(len(mesh.vertices), segments+1)
                self.assertEqual(len(mesh.faces), 2*(segments-1))
                if segments == 4:
                    self.assertAlmostEqual(mesh.area, 2*(1+math.sqrt(3)))
                    self.assertAlmostEqual(mesh.volume, 2/3)
                else:
                    self.assertAlmostEqual(mesh.area, (1+math.sqrt(2))*math.pi, places=int(math.log10(segments)))
                    self.assertAlmostEqual(mesh.volume, math.pi/3, places=int(math.log10(segments)))


class TestTransform(utils.TestCase):
    def tearDown(self):
        Model.flush_caches(0, 0)

    def assertTransform(self, base_model, matrix):
        transformed_model = base_model.transform(matrix)
        self.assertEqual(transformed_model.name, f'{base_model.name}@{hex(matrix.hash(False))[2:]}')
        base_mesh = base_model.get_trimesh()
        transformed_mesh = transformed_model.get_trimesh()
        self.assertEqual(len(base_mesh.vertices), len(transformed_mesh.vertices))
        self.assertEqual(base_mesh.faces.tolist(), transformed_mesh.faces.tolist())
        for vertex1, vertex2 in zip(base_mesh.vertices.tolist(), transformed_mesh.vertices.tolist()):
            self.assertAllAlmostEqual(vertex2, matrix @ vertex1, places=6)
        base_mesh = base_model.get_manifold().to_mesh()
        transformed_mesh = transformed_model.get_manifold().to_mesh()
        self.assertEqual(len(base_mesh.vert_properties), len(transformed_mesh.vert_properties))
        self.assertEqual(base_mesh.tri_verts.tolist(), transformed_mesh.tri_verts.tolist())
        for vertex1, vertex2 in zip(base_mesh.vert_properties.tolist(), transformed_mesh.vert_properties.tolist()):
            self.assertAllAlmostEqual(vertex2, matrix @ vertex1, places=6)

    def test_translate(self):
        self.assertTransform(Model.cylinder(), Matrix44.translate((1, 2, 3)))

    def test_rotate(self):
        self.assertTransform(Model.cylinder(), Matrix44.rotate(1/3))

    def test_scale(self):
        self.assertTransform(Model.cylinder(), Matrix44.scale((2, 3, 4)))


class TestCache(utils.TestCase):
    def tearDown(self):
        Model.flush_caches(0, 0)

    def test_cached_trimesh(self):
        model = Model.box()
        bounds = model.get_bounds()
        mesh = model.get_trimesh()
        manifold = model.get_manifold()
        self.assertIs(model.get_bounds(), bounds)
        self.assertIs(model.get_trimesh(), mesh)
        self.assertIs(model.get_manifold(), manifold)

    def test_invalidate(self):
        model = Model.box()
        bounds = model.get_bounds()
        mesh = model.get_trimesh()
        manifold = model.get_manifold()
        self.assertIs(model.get_bounds(), bounds)
        self.assertIs(model.get_trimesh(), mesh)
        self.assertIs(model.get_manifold(), manifold)
        model.invalidate()
        self.assertIsNot(model.get_bounds(), bounds)
        self.assertIsNot(model.get_trimesh(), mesh)
        self.assertIsNot(model.get_manifold(), manifold)

    def test_invalidate_dependency(self):
        box = Model.box()
        transformed_box = box.transform(Matrix44.scale(2))
        self.assertIsNot(transformed_box, box)
        self.assertIn(transformed_box, box.dependents)
        bounds = transformed_box.get_bounds()
        mesh = transformed_box.get_trimesh()
        manifold = transformed_box.get_manifold()
        self.assertIs(transformed_box.get_bounds(), bounds)
        self.assertIs(transformed_box.get_trimesh(), mesh)
        self.assertIs(transformed_box.get_manifold(), manifold)
        box.invalidate()
        self.assertIsNot(transformed_box.get_bounds(), bounds)
        self.assertIsNot(transformed_box.get_trimesh(), mesh)
        self.assertIsNot(transformed_box.get_manifold(), manifold)


class TestStructuring(utils.TestCase):
    def setUp(self):
        self.M = Matrix44.translate(1)
        self.M_hash = hex(self.M.hash(False))[2:]
        self.M2 = Matrix44.translate(2)
        self.M2_hash = hex(self.M2.hash(False))[2:]
        self.P = Vector((0, 0, 0))
        self.N = Vector((1, 0, 0))
        self.PN_hash = hex(self.P.hash(False) ^ self.N.hash(False))[2:]

    def tearDown(self):
        Model.flush_caches(0, 0)

    def test_flatten(self):
        self.assertEqual(Model.box().flatten().name, 'flatten(!box)')
        self.assertFalse(Model.box().flatten().is_manifold())
        self.assertIs(Model.box().flatten().get_manifold(), Model.box().get_manifold())
        self.assertFalse(Model.union(Model.box(), Model.sphere()).flatten().is_manifold())
        self.assertEqual(Model.box().flatten().flatten().name, 'flatten(!box)')
        self.assertEqual(Model.box().flatten().invert().name, 'invert(flatten(!box))')
        self.assertEqual(Model.box().flatten().repair().name, 'repair(flatten(!box))')
        self.assertEqual(Model.box().flatten().snap_edges().name, 'flatten(!box)')
        self.assertEqual(Model.box().flatten().transform(self.M).name, f'flatten(!box)@{self.M_hash}')
        self.assertEqual(Model.box().flatten().uv_remap('sphere').name, 'uv_remap(flatten(!box), sphere)')
        self.assertEqual(Model.box().flatten().trim(self.P, self.N).name, f'trim(flatten(!box), {self.PN_hash})')

    def test_invert(self):
        self.assertEqual(Model.box().invert().name, 'invert(!box)')
        self.assertFalse(Model.box().invert().is_manifold())
        self.assertIs(Model.box().invert().get_manifold(), Model.box().get_manifold())
        self.assertFalse(Model.union(Model.box(), Model.sphere()).invert().is_manifold())
        self.assertEqual(Model.box().invert().flatten().name, 'flatten(invert(!box))')
        self.assertEqual(Model.box().invert().invert().name, '!box')
        self.assertEqual(Model.box().invert().repair().name, 'invert(repair(!box))')
        self.assertEqual(Model.box().invert().snap_edges().name, 'invert(snap_edges(!box))')
        self.assertEqual(Model.box().invert().transform(self.M).name, f'invert(!box@{self.M_hash})')
        self.assertEqual(Model.box().invert().uv_remap('sphere').name, 'uv_remap(invert(!box), sphere)')
        self.assertEqual(Model.box().invert().trim(self.P, self.N).name, f'trim(invert(!box), {self.PN_hash})')

    def test_repair(self):
        self.assertEqual(Model.box().repair().name, 'repair(!box)')
        self.assertFalse(Model.box().repair().is_manifold())
        self.assertIsNot(Model.box().repair().get_manifold(), Model.box().get_manifold())
        self.assertEqual(Model.box().repair().flatten().name, 'flatten(repair(!box))')
        self.assertEqual(Model.box().repair().invert().name, 'invert(repair(!box))')
        self.assertEqual(Model.box().repair().repair().name, 'repair(!box)')
        self.assertEqual(Model.box().repair().snap_edges().name, 'snap_edges(repair(!box))')
        self.assertEqual(Model.box().repair().transform(self.M).name, f'repair(!box)@{self.M_hash}')
        self.assertEqual(Model.box().repair().uv_remap('sphere').name, 'uv_remap(repair(!box), sphere)')
        self.assertEqual(Model.box().repair().trim(self.P, self.N).name, f'trim(repair(!box), {self.PN_hash})')

    def test_snap_edges(self):
        self.assertEqual(Model.box().snap_edges(0).name, 'flatten(!box)')
        self.assertFalse(Model.box().snap_edges().is_manifold())
        self.assertIsNot(Model.box().snap_edges().get_manifold(), Model.box().get_manifold())
        self.assertFalse(Model.union(Model.box(), Model.sphere()).snap_edges().is_manifold())
        self.assertEqual(Model.box().snap_edges().name, 'snap_edges(!box)')
        self.assertEqual(Model.box().snap_edges(0.05).name, 'snap_edges(!box)')
        self.assertEqual(Model.box().snap_edges(0.25).name, 'snap_edges(!box, 0.25)')
        self.assertEqual(Model.box().snap_edges().flatten().name, 'flatten(!box)')
        self.assertEqual(Model.box().snap_edges().invert().name, 'invert(snap_edges(!box))')
        self.assertEqual(Model.box().snap_edges().repair().name, 'snap_edges(repair(!box))')
        self.assertEqual(Model.box().snap_edges().snap_edges().name, 'snap_edges(!box)')
        self.assertEqual(Model.box().snap_edges().transform(self.M).name, f'snap_edges(!box@{self.M_hash})')
        self.assertEqual(Model.box().snap_edges().uv_remap('sphere').name, 'uv_remap(snap_edges(!box), sphere)')
        self.assertEqual(Model.box().snap_edges().trim(self.P, self.N).name, f'trim(!box, {self.PN_hash})')

    def test_transform(self):
        self.assertEqual(Model.box().transform(Matrix44()).name, '!box')
        self.assertFalse(Model.box().transform(self.M).is_manifold())
        self.assertTrue(Model.union(Model.box(), Model.sphere()).transform(self.M).is_manifold())
        self.assertEqual(Model.box().transform(self.M).name, f'!box@{self.M_hash}')
        self.assertEqual(Model.box().transform(self.M).flatten().name, f'flatten(!box@{self.M_hash})')
        self.assertEqual(Model.box().transform(self.M).invert().name, f'invert(!box@{self.M_hash})')
        self.assertEqual(Model.box().transform(self.M).repair().name, f'repair(!box)@{self.M_hash}')
        self.assertEqual(Model.box().transform(self.M).snap_edges().name, f'snap_edges(!box@{self.M_hash})')
        self.assertEqual(Model.box().transform(self.M).transform(self.M).name, f'!box@{self.M2_hash}')
        self.assertEqual(Model.box().transform(self.M).uv_remap('sphere').name, f'uv_remap(!box@{self.M_hash}, sphere)')
        self.assertEqual(Model.box().transform(self.M).trim(self.P, self.N).name, f'trim(!box@{self.M_hash}, {self.PN_hash})')

    def test_uvremap(self):
        self.assertEqual(Model.box().uv_remap('sphere').name, 'uv_remap(!box, sphere)')
        self.assertFalse(Model.box().uv_remap('sphere').is_manifold())
        self.assertIs(Model.box().uv_remap('sphere').get_manifold(), Model.box().get_manifold())
        self.assertFalse(Model.union(Model.box(), Model.sphere()).uv_remap('sphere').is_manifold())
        self.assertEqual(Model.box().uv_remap('sphere').flatten().name, 'flatten(uv_remap(!box, sphere))')
        self.assertEqual(Model.box().uv_remap('sphere').invert().name, 'invert(uv_remap(!box, sphere))')
        self.assertEqual(Model.box().uv_remap('sphere').repair().name, 'uv_remap(repair(!box), sphere)')
        self.assertEqual(Model.box().uv_remap('sphere').snap_edges().name, 'snap_edges(uv_remap(!box, sphere))')
        self.assertEqual(Model.box().uv_remap('sphere').transform(self.M).name, f'uv_remap(!box, sphere)@{self.M_hash}')
        self.assertEqual(Model.box().uv_remap('test').uv_remap('sphere').name, 'uv_remap(!box, sphere)')
        self.assertEqual(Model.box().uv_remap('sphere').trim(self.P, self.N).name, f'trim(uv_remap(!box, sphere), {self.PN_hash})')

    def test_trim(self):
        self.assertEqual(Model.box().trim(self.P, self.N).name, f'trim(!box, {self.PN_hash})')
        self.assertTrue(Model.box().trim(self.P, self.N).is_manifold())
        self.assertEqual(Model.box().trim(self.P, self.N).flatten().name, f'flatten(trim(!box, {self.PN_hash}))')
        self.assertEqual(Model.box().trim(self.P, self.N).invert().name, f'invert(trim(!box, {self.PN_hash}))')
        self.assertEqual(Model.box().trim(self.P, self.N).repair().name, f'trim(repair(!box), {self.PN_hash})')
        self.assertEqual(Model.box().trim(self.P, self.N).snap_edges().name, f'snap_edges(trim(!box, {self.PN_hash}))')
        MPN_hash = hex((self.M @ self.P).hash(False) ^ ((self.M.matrix33_cofactor() @ self.N).normalize()).hash(False))[2:]
        self.assertEqual(Model.box().trim(self.P, self.N).transform(self.M).name, f'trim(!box@{self.M_hash}, {MPN_hash})')
        self.assertEqual(Model.box().trim(self.P, self.N).uv_remap('sphere').name, f'uv_remap(trim(!box, {self.PN_hash}), sphere)')
        self.assertEqual(Model.box().trim(self.P, self.N).trim(self.P, self.N).name, f'trim(trim(!box, {self.PN_hash}), {self.PN_hash})')

    def test_union(self):
        self.assertIsNone(Model.union())
        self.assertEqual(Model.union(Model.box()).name, '!box')
        self.assertEqual(Model.union(Model.box(), Model.box()).name, '!box')
        self.assertEqual(Model.union(Model.box(), Model.sphere()).name, 'union(!box, !sphere)')
        self.assertEqual(Model.union(Model.box(), Model.sphere(), Model.box()).name, 'union(!box, !sphere)')
        self.assertEqual(Model.union(Model.box(), Model.union(Model.sphere(), Model.cylinder())).name, 'union(!box, !sphere, !cylinder)')
        self.assertTrue(Model.union(Model.box(), Model.sphere()).is_manifold())
        self.assertEqual(Model.union(Model.box(), Model.sphere()).flatten().name, 'flatten(union(!box, !sphere))')
        self.assertEqual(Model.union(Model.box(), Model.sphere()).invert().name, 'invert(union(!box, !sphere))')
        self.assertEqual(Model.union(Model.box(), Model.sphere()).repair().name, 'union(!box, !sphere)')
        self.assertEqual(Model.union(Model.box(), Model.sphere()).snap_edges().name, 'snap_edges(union(!box, !sphere))')
        self.assertEqual(Model.union(Model.box(), Model.sphere()).transform(self.M).name, f'union(!box@{self.M_hash}, !sphere@{self.M_hash})')
        self.assertEqual(Model.union(Model.box(), Model.sphere()).uv_remap('sphere').name, 'uv_remap(union(!box, !sphere), sphere)')
        self.assertEqual(Model.union(Model.box(), Model.sphere()).trim(self.P, self.N).name, f'trim(union(!box, !sphere), {self.PN_hash})')

    def test_intersect(self):
        self.assertIsNone(Model.intersect())
        self.assertEqual(Model.intersect(Model.box()).name, '!box')
        self.assertEqual(Model.intersect(Model.box(), Model.box()).name, '!box')
        self.assertEqual(Model.intersect(Model.box(), Model.sphere()).name, 'intersect(!box, !sphere)')
        self.assertEqual(Model.intersect(Model.box(), Model.sphere(), Model.box()).name, 'intersect(!box, !sphere)')
        self.assertEqual(Model.intersect(Model.box(), Model.intersect(Model.sphere(), Model.cylinder())).name, 'intersect(!box, !sphere, !cylinder)')
        self.assertTrue(Model.intersect(Model.box(), Model.sphere()).is_manifold())
        self.assertEqual(Model.intersect(Model.box(), Model.sphere()).flatten().name, 'flatten(intersect(!box, !sphere))')
        self.assertEqual(Model.intersect(Model.box(), Model.sphere()).invert().name, 'invert(intersect(!box, !sphere))')
        self.assertEqual(Model.intersect(Model.box(), Model.sphere()).repair().name, 'intersect(!box, !sphere)')
        self.assertEqual(Model.intersect(Model.box(), Model.sphere()).snap_edges().name, 'snap_edges(intersect(!box, !sphere))')
        self.assertEqual(Model.intersect(Model.box(), Model.sphere()).transform(self.M).name, f'intersect(!box@{self.M_hash}, !sphere@{self.M_hash})')
        self.assertEqual(Model.intersect(Model.box(), Model.sphere()).uv_remap('sphere').name, 'uv_remap(intersect(!box, !sphere), sphere)')
        self.assertEqual(Model.intersect(Model.box(), Model.sphere()).trim(self.P, self.N).name, f'intersect(trim(!box, {self.PN_hash}), !sphere)')

    def test_difference(self):
        self.assertIsNone(Model.difference())
        self.assertEqual(Model.difference(Model.box()).name, '!box')
        self.assertIsNone(Model.difference(Model.box(), Model.box()))
        self.assertEqual(Model.difference(Model.box(), Model.sphere()).name, 'difference(!box, !sphere)')
        self.assertIsNone(Model.difference(Model.box(), Model.sphere(), Model.box()))
        self.assertEqual(Model.difference(Model.box(), Model.difference(Model.sphere(), Model.cylinder())).name,
                         'difference(!box, difference(!sphere, !cylinder))')
        self.assertTrue(Model.difference(Model.box(), Model.sphere()).is_manifold())
        self.assertEqual(Model.difference(Model.box(), Model.sphere()).flatten().name, 'flatten(difference(!box, !sphere))')
        self.assertEqual(Model.difference(Model.box(), Model.sphere()).invert().name, 'invert(difference(!box, !sphere))')
        self.assertEqual(Model.difference(Model.box(), Model.sphere()).repair().name, 'difference(!box, !sphere)')
        self.assertEqual(Model.difference(Model.box(), Model.sphere()).snap_edges().name, 'snap_edges(difference(!box, !sphere))')
        self.assertEqual(Model.difference(Model.box(), Model.sphere()).transform(self.M).name, f'difference(!box@{self.M_hash}, !sphere@{self.M_hash})')
        self.assertEqual(Model.difference(Model.box(), Model.sphere()).uv_remap('sphere').name, 'uv_remap(difference(!box, !sphere), sphere)')
        self.assertEqual(Model.difference(Model.box(), Model.sphere()).trim(self.P, self.N).name, f'difference(trim(!box, {self.PN_hash}), !sphere)')


class TestUnaryOperations(utils.TestCase):
    def tearDown(self):
        Model.flush_caches(0, 0)

    def test_uv_remap_sphere(self):
        model = Model.box()
        mesh = model.uv_remap('sphere').get_trimesh()
        for (x, y, z), uv in zip(mesh.vertices, mesh.visual.uv):
            u = math.atan2(y, x) / (2*math.pi) % 1
            r = math.sqrt(x*x + y*y)
            v = (math.atan2(z, r) / math.pi + 0.5) % 1
            self.assertAllAlmostEqual(uv, [u, v])

    def test_flatten(self):
        model = Model.sphere()
        nrows = DefaultSegments // 4
        mesh = model.get_trimesh()
        self.assertEqual(len(mesh.vertices), 4*(nrows+1)*(nrows+2))
        self.assertEqual(len(mesh.faces), 8*nrows*nrows)
        self.assertAlmostEqual(mesh.area, 4*math.pi, places=1)
        self.assertAlmostEqual(mesh.volume, 4/3*math.pi, places=1)
        flat_model = model.flatten()
        self.assertIsNot(flat_model, model)
        mesh = flat_model.get_trimesh()
        self.assertEqual(len(mesh.vertices), 3*8*nrows*nrows)
        self.assertEqual(len(mesh.faces), 8*nrows*nrows)
        self.assertAlmostEqual(mesh.area, 4*math.pi, places=1)
        self.assertAlmostEqual(mesh.volume, 4/3*math.pi, places=1)


class TestTrim(utils.TestCase):
    def tearDown(self):
        Model.flush_caches(0, 0)

    def test_bad_arguments(self):
        self.assertIsNone(Model.box().trim(None, [1, 0, 0]))
        self.assertIsNone(Model.box().trim(0, None))

    def test_trim_sphere_to_box(self):
        model = Model.sphere()
        for x in (-1, 1):
            model = model.trim((x/2, 0, 0), (x, 0, 0))
        for y in (-1, 1):
            model = model.trim((0, y/2, 0), (0, y, 0))
        for z in (-1, 1):
            model = model.trim((0, 0, z/2), (0, 0, z))
        self.assertTrue(model.is_manifold())
        mesh = model.get_trimesh()
        self.assertEqual(mesh.bounds.tolist(), [[-0.5, -0.5, -0.5], [0.5, 0.5, 0.5]])
        self.assertEqual(len(mesh.vertices), 8)
        self.assertEqual(len(mesh.faces), 12)
        self.assertAlmostEqual(mesh.area, 6)
        self.assertAlmostEqual(mesh.volume, 1)

    def test_trim_to_nothing(self):
        model = Model.box()
        model = model.trim((0, 0, -1), (0, 0, 1))
        self.assertTrue(model.is_manifold())
        with unittest.mock.patch('flitter.render.window.models.logger') as mock_logger:
            manifold = model.get_manifold()
            mock_logger.warning.assert_called_with("Result of operation was empty mesh: {}", model.name)
        self.assertIsNone(manifold)
        self.assertIsNone(model.get_trimesh())


class TestBoolean(utils.TestCase):
    Accuracy = 0.01

    def setUp(self):
        self.nested_box_models = [Model.box().transform(Matrix44.scale(2)), Model.box()]
        self.cross_models = [Model.box().transform(Matrix44.scale([3, 1, 1])),
                             Model.box().transform(Matrix44.scale([1, 3, 1])),
                             Model.box().transform(Matrix44.scale([1, 1, 3]))]
        self.capsule_models = [Model.cylinder().transform(Matrix44.scale([1, 1, 2])),
                               Model.sphere().transform(Matrix44.translate([0, 0, 1])),
                               Model.sphere().transform(Matrix44.translate([0, 0, -1]))]

    def tearDown(self):
        Model.flush_caches(0, 0)

    def wrap_model(self, model):
        return model

    def wrap_name(self, name):
        return name

    def test_ignored_none(self):
        self.assertEqual(self.wrap_model(Model.union(Model.box(), None, Model.sphere())).name,
                         self.wrap_name('union(!box, !sphere)'))
        self.assertEqual(self.wrap_model(Model.intersect(Model.box(), None, Model.sphere())).name,
                         self.wrap_name('intersect(!box, !sphere)'))
        self.assertEqual(self.wrap_model(Model.difference(Model.box(), None, Model.sphere())).name,
                         self.wrap_name('difference(!box, !sphere)'))

    def test_is_manifold(self):
        self.assertTrue(self.wrap_model(Model.union(Model.box(), Model.sphere())).is_manifold())
        self.assertTrue(self.wrap_model(Model.intersect(Model.box(), Model.sphere())).is_manifold())
        self.assertTrue(self.wrap_model(Model.difference(Model.box(), Model.sphere())).is_manifold())

    def test_nested_box_union(self):
        model = self.wrap_model(Model.union(*self.nested_box_models))
        mesh = model.get_trimesh()
        self.assertAlmostEqual(mesh.area, 24)
        self.assertAlmostEqual(mesh.volume, 8)

    def test_nested_box_intersect(self):
        model = self.wrap_model(Model.intersect(*self.nested_box_models))
        mesh = model.get_trimesh()
        self.assertAlmostEqual(mesh.area, 6)
        self.assertAlmostEqual(mesh.volume, 1)

    def test_nested_box_difference(self):
        model = self.wrap_model(Model.difference(*self.nested_box_models))
        mesh = model.get_trimesh()
        self.assertAlmostEqual(mesh.area, 30)
        self.assertAlmostEqual(mesh.volume, 7)

    def test_nested_box_reversed_difference(self):
        model = self.wrap_model(Model.difference(*reversed(self.nested_box_models)))
        with unittest.mock.patch('flitter.render.window.models.logger') as mock_logger:
            mesh = model.get_trimesh()
            mock_logger.warning.assert_called_with("Result of operation was empty mesh: {}", model.name)
        self.assertIsNone(mesh)

    def test_cross_union(self):
        model = self.wrap_model(Model.union(*self.cross_models))
        mesh = model.get_trimesh()
        self.assertAlmostEqual(mesh.area, 30, places=None, delta=self.Accuracy)
        self.assertAlmostEqual(mesh.volume, 7, places=None, delta=self.Accuracy)

    def test_cross_intersect(self):
        model = self.wrap_model(Model.intersect(*self.cross_models))
        mesh = model.get_trimesh()
        self.assertAlmostEqual(mesh.area, 6, places=None, delta=self.Accuracy)
        self.assertAlmostEqual(mesh.volume, 1, places=None, delta=self.Accuracy)

    def test_cross_difference(self):
        model = self.wrap_model(Model.difference(*self.cross_models))
        mesh = model.get_trimesh()
        self.assertAlmostEqual(mesh.area, 12, places=None, delta=self.Accuracy)
        self.assertAlmostEqual(mesh.volume, 2, places=None, delta=self.Accuracy)

    def test_capsule_union(self):
        model = self.wrap_model(Model.union(*self.capsule_models))
        mesh = model.get_trimesh()
        self.assertAlmostEqual(mesh.area, 4*math.pi + 4*math.pi, places=None, delta=self.Accuracy*5)
        self.assertAlmostEqual(mesh.volume, 2*math.pi + 4/3*math.pi, places=None, delta=self.Accuracy*5)

    def test_capsule_intersect(self):
        model = self.wrap_model(Model.intersect(*self.capsule_models))
        with unittest.mock.patch('flitter.render.window.models.logger') as mock_logger:
            mesh = model.get_trimesh()
            mock_logger.warning.assert_called_with("Result of operation was empty mesh: {}", model.name)
        self.assertIsNone(mesh)

    def test_capsule_difference(self):
        model = self.wrap_model(Model.difference(*self.capsule_models))
        mesh = model.get_trimesh()
        self.assertAlmostEqual(mesh.area, 4*math.pi + 4*math.pi, places=None, delta=self.Accuracy*30)
        self.assertAlmostEqual(mesh.volume, 2*math.pi - 4/3*math.pi, places=None, delta=self.Accuracy*5)


class TestSdfBoolean(TestBoolean):
    Accuracy = 0.2

    def wrap_model(self, model):
        return Model.sdf(None, model, (-3, -3, -3), (3, 3, 3), 0.1)

    def wrap_name(self, name):
        return f'sdf({name}, -3;-3;-3, 3;3;3, 0.1)'


class TestSdfTrim(utils.TestCase):
    def tearDown(self):
        Model.flush_caches(0, 0)

    def test_is_manifold(self):
        self.assertTrue(Model.sphere().trim((0, 0, 0), (0, 0, 1)).is_manifold())

    def test_trim_sphere_to_box(self):
        model = Model.sphere()
        for x in (-1, 1):
            model = model.trim((x/2, 0, 0), (x, 0, 0))
        for y in (-1, 1):
            model = model.trim((0, y/2, 0), (0, y, 0))
        for z in (-1, 1):
            model = model.trim((0, 0, z/2), (0, 0, z))
        model = Model.sdf(None, model, (-1, -1, -1), (1, 1, 1), 0.05)
        mesh = model.get_trimesh()
        self.assertEqual(mesh.bounds.tolist(), [[-0.5, -0.5, -0.5], [0.5, 0.5, 0.5]])
        self.assertAlmostEqual(mesh.area, 6)
        self.assertAlmostEqual(mesh.volume, 1)

    def test_trim_to_nothing(self):
        model = Model.box()
        model = model.trim((0, 0, -1), (0, 0, 1))
        model = Model.sdf(None, model, (-1, -1, -1), (1, 1, 1), 0.05)
        with unittest.mock.patch('flitter.render.window.models.logger') as mock_logger:
            manifold = model.get_manifold()
            mock_logger.warning.assert_called_with("Result of operation was empty mesh: {}", model.name)
        self.assertIsNone(manifold)
        self.assertIsNone(model.get_trimesh())


class TestBuffers(utils.TestCase):
    def tearDown(self):
        Model.flush_caches(0, 0)

    def test_box_get_buffers(self):
        model = Model.box()
        glctx = unittest.mock.Mock()
        objects = {}
        buffers = model.get_buffers(glctx, objects)
        self.assertIn(model.id, objects)
        self.assertIs(objects[model.id], buffers)
        vertex_data, = glctx.buffer.mock_calls[0].args
        self.assertEqual(vertex_data.dtype.name, 'float32')
        self.assertEqual(vertex_data.shape, (24, 8))
        index_data, = glctx.buffer.mock_calls[1].args
        self.assertEqual(index_data.dtype.name, 'int32')
        self.assertEqual(index_data.shape, (12, 3))
        model.invalidate()
        self.assertNotIn(model.id, objects)
